#  -*- coding: utf-8 -*-
"""Tests for sksrurgerycalibration pivot calibration"""
from glob import glob
from random import seed
import numpy as np
import pytest
import sksurgerycalibration.algorithms.pivot as p


def test_empty_matrices():
    """Throws a type error if empty matrices are None"""

    with pytest.raises(TypeError):
        p.pivot_calibration(None)


def test_rank_lt_six():
    """Throw a value error if matrix rank is less than 6?"""
    with pytest.raises(ValueError):
        file_names = glob('tests/data/PivotCalibration/1378476417807806000.txt')
        arrays = [np.loadtxt(f) for f in file_names]
        matrices = np.concatenate(arrays)
        number_of_matrices = int(matrices.size/16)
        matrices = matrices.reshape((number_of_matrices, 4, 4))
        p.pivot_calibration_aos(matrices)


def test_unkown_method():
    """Throw value error if config set for unknown method"""
    config = {"method" : "not implemented"}
    matrix = (np.arange(2, 18, dtype=float).reshape((1, 4, 4)))
    with pytest.raises(ValueError):
        p.pivot_calibration(matrix, config)

def test_four_columns_matrices4x4():
    """Throw a value error if matrix is not 4 column"""

    with pytest.raises(ValueError):
        p.pivot_calibration(np.arange(2, 14, dtype=float).reshape((1, 4, 3)))


def test_four_rows_matrices4x4():
    """Throw a value error if matrix is not 4 rows"""

    with pytest.raises(ValueError):
        p.pivot_calibration(np.arange(2, 14, dtype=float).reshape((1, 3, 4)))


def test_return_value():
    """A regression test using some recorded data"""

    file_names = glob('tests/data/PivotCalibration/*')
    arrays = [np.loadtxt(f) for f in file_names]
    matrices = np.concatenate(arrays)
    number_of_matrices = int(matrices.size/16)
    matrices = matrices.reshape((number_of_matrices, 4, 4))
    config = {"method" : "aos"}
    pointer_offset, pivot_point, residual_error = \
        p.pivot_calibration(matrices, config)
    assert round(residual_error, 3) == 1.761
    assert round(pointer_offset[0, 0], 3) == -14.473
    assert round(pointer_offset[1, 0], 3) == 394.634
    assert round(pointer_offset[2, 0], 3) == -7.407
    assert round(pivot_point[0, 0], 3) == -804.742
    assert round(pivot_point[1, 0], 3) == -85.474
    assert round(pivot_point[2, 0], 3) == -2112.131


def test_pivot_with_ransac():
    """Tests that pivot with ransac runs"""
    #seed the random number generator. Seeding
    #with 0 leads to one failed pivot calibration (rank < 6), so we
    #hit lines 127-129
    seed(0)

    file_names = glob('tests/data/PivotCalibration/*')
    arrays = [np.loadtxt(f) for f in file_names]
    matrices = np.concatenate(arrays)
    number_of_matrices = int(matrices.size/16)
    matrices = matrices.reshape((number_of_matrices, 4, 4))
    _, _, residual_1 = p.pivot_calibration(matrices)
    _, _, residual_2 = p.pivot_calibration_with_ransac(matrices, 10, 4, 0.25)
    assert residual_2 < residual_1
    _, _, _ = p.pivot_calibration_with_ransac(matrices,
                                              10, 4, 0.25,
                                              early_exit=True)
    #tests for the value checkers at the start of RANSAC
    with pytest.raises(ValueError):
        _, _, _ = p.pivot_calibration_with_ransac(None, 0, None, None)

    with pytest.raises(ValueError):
        _, _, _ = p.pivot_calibration_with_ransac(None, 2, -1.0, None)

    with pytest.raises(ValueError):
        _, _, _ = p.pivot_calibration_with_ransac(None, 2, 1.0, 1.1)

    with pytest.raises(TypeError):
        _, _, _ = p.pivot_calibration_with_ransac(None, 2, 1.0, 0.8)

    #with consensus threshold set to 1.0, we get a value error
    #as no best model is found.
    with pytest.raises(ValueError):
        _, _, _ = p.pivot_calibration_with_ransac(matrices,
                                                  10, 4, 1.0,
                                                  early_exit=True)


def test_pivot_with_sphere_fit():
    """Tests pivot calibration with sphere fitting"""
    config = {"method" : "sphere_fitting"}
    file_names = glob('tests/data/PivotCalibration/*')
    arrays = [np.loadtxt(f) for f in file_names]
    matrices = np.concatenate(arrays)
    number_of_matrices = int(matrices.size/16)
    matrices = matrices.reshape((number_of_matrices, 4, 4))
    _, _, residual_error = p.pivot_calibration(matrices, config)

    #do a regression test on the residual error
    assert round(residual_error, 3) == 2.346


def test_replace_small_values():
    """Tests for small values replacement"""
    list_in = [0.2, 0.6, 0.0099, 0.56]

    rank = p._replace_small_values(list_in) #pylint: disable=protected-access

    assert rank == 3
    assert list_in[2] == 0

    rank = p._replace_small_values( #pylint: disable=protected-access
        list_in,
        threshold=0.3,
        replacement_value=-1.0)

    assert rank == 2
    assert list_in[0] == -1.0
    assert list_in[2] == -1.0


def _rms(*values):
    """
    Works out RMS error, so we can test against it.
    """
    count = len(values)
    square_sum = 0.0
    for value in values:
        square_sum += value * value

    mean_square_sum = square_sum / count
    return np.sqrt(mean_square_sum)

def test_residual_error0():
    """
    Test that residual error returns a correct value
    """

    pivot_point = [0.0, 0.0, 0.0]
    pointer_offset = [-100.0, 0.0, 0.0]

    tracker_mat = np.array([[[1.0, 0.0, 0.0, 110.0],
                             [0.0, 1.0, 0.0, 0.0],
                             [0.0, 0.0, 1.0, 0.0],
                             [0.0, 0.0, 0.0, 1.0]]
                           ])
    #error is 10.0 in one direction and zero in the others
    expected_value = _rms(10.0, 0.0, 0.0)
    residual_error = p._residual_error( #pylint: disable=protected-access
        tracker_mat,
        pointer_offset, pivot_point)
    assert residual_error == expected_value


def test_residual_error1():
    """
    Test that residual error returns a correct value
    """

    pivot_point = [0.0, 0.0, 0.0]
    pointer_offset = [-100.0, 0.0, 0.0]

    tracker_mat = np.array([[[1.0, 0.0, 0.0, 100.0],
                             [0.0, 1.0, 0.0, 0.0],
                             [0.0, 0.0, 1.0, 0.0],
                             [0.0, 0.0, 0.0, 1.0]]
                           ])

    #error in all three directions is zero
    expected_value = _rms(0.0, 0.0, 0.0)
    residual_error = p._residual_error( #pylint: disable=protected-access
        tracker_mat,
        pointer_offset, pivot_point)
    assert residual_error == expected_value


def test_residual_error2():
    """
    Test that residual error returns a correct value
    """

    pivot_point = [0.0, 0.0, 0.0]
    pointer_offset = [-100.0, 0.0, 0.0]


    tracker_mat = np.array([[[1.0, 0.0, 0.0, 110.0],
                             [0.0, 1.0, 0.0, 0.0],
                             [0.0, 0.0, 1.0, 0.0],
                             [0.0, 0.0, 0.0, 1.0]],

                            [[1.0, 0.0, 0.0, 90.0],
                             [0.0, 1.0, 0.0, 0.0],
                             [0.0, 0.0, 1.0, 0.0],
                             [0.0, 0.0, 0.0, 1.0]],
                           ])

    #error in all 2 of 6 directions is 10.0
    expected_value = _rms(10.0, 0.0, 0.0, -10.0, 0.0, 0.0)
    residual_error = p._residual_error( #pylint: disable=protected-access
        tracker_mat,
        pointer_offset, pivot_point)
    assert residual_error == expected_value


def test_residual_error3():
    """
    Test that residual error returns a correct value
    """

    pivot_point = [0.0, 0.0, 0.0]
    pointer_offset = [-100.0, 0.0, 0.0]

    tracker_mat = np.array([[[1.0, 0.0, 0.0, 110.0],
                             [0.0, 1.0, 0.0, 10.0],
                             [0.0, 0.0, 1.0, -10.0],
                             [0.0, 0.0, 0.0, 1.0]]
                           ])

    #error in all three directions is 10.0
    expected_value = _rms(10.0, 10.0, -10.0)
    residual_error = p._residual_error( #pylint: disable=protected-access
        tracker_mat,
        pointer_offset, pivot_point)
    assert residual_error == expected_value



# -*- coding: utf-8 -*-

# pylint: disable=unused-import, superfluous-parens, line-too-long, missing-module-docstring, unused-variable, missing-function-docstring, invalid-name

import glob
import pytest
import numpy as np
import sksurgerycalibration.video.video_calibration_utils as vu
import sksurgerycalibration.video.video_calibration_wrapper as vc


def test_mono_left_video_calibration():

    image_points = []
    object_points = []

    model = np.loadtxt('tests/data/laparoscope_calibration/chessboard_14_10_3.txt')

    files = glob.glob('tests/data/laparoscope_calibration/left/*.txt')
    for file in files:
        points = np.loadtxt(file)
        image_points.append(vu.convert_numpy2d_to_opencv(points))
        object_points.append(vu.convert_numpy3d_to_opencv(model))

    retval, camera_matrix, dist_coeffs, rvecs, tvecs = \
        vc.mono_video_calibration(object_points, image_points, (1920, 1080))

    assert(np.abs(retval - 0.57759896) < 0.000001)


def load_first_stereo_data():

    left_image_points = []
    right_image_points = []
    object_points = []

    model = np.loadtxt('tests/data/laparoscope_calibration/chessboard_14_10_3.txt')

    files = glob.glob('tests/data/laparoscope_calibration/left/*.txt')
    files.sort()
    for file in files:
        points = np.loadtxt(file)
        left_image_points.append(vu.convert_numpy2d_to_opencv(points))
        object_points.append(vu.convert_numpy3d_to_opencv(model))

    files = glob.glob('tests/data/laparoscope_calibration/right/*.txt')
    files.sort()
    for file in files:
        points = np.loadtxt(file)
        right_image_points.append(vu.convert_numpy2d_to_opencv(points))

    # Generates, array of arrays
    ids = []
    for i in range(9):
        ids.append(np.asarray(range(140)))

    return object_points, \
           left_image_points, \
           right_image_points, \
           ids


def test_stereo_video_calibration():

    object_points, left_image_points, right_image_points, ids = \
        load_first_stereo_data()

    s_reproj, s_recon, \
        l_c, l_d, left_rvecs, left_tvecs, \
        r_c, r_d, right_rvecs, right_tvecs, \
        l2r_r, l2r_t, \
        essential, fundamental = \
        vc.stereo_video_calibration(ids,
                                    object_points,
                                    left_image_points,
                                    ids,
                                    object_points,
                                    right_image_points,
                                    (1920, 1080))

    assert (np.abs(s_reproj - 0.63022577) < 0.000001)
    assert (np.abs(s_recon - 1.64274596) < 0.000001)

# def test_experimental_mono_stereo_calib():
#
#   This isn't worth the bother. Takes 5 minutes to run.
#
#    image_points = []
#    object_points = []
#
#    model = np.loadtxt('tests/data/laparoscope_calibration/chessboard_14_10_3.txt')
#
#    files = glob.glob('tests/data/laparoscope_calibration/left/*.txt')
#    for file in files:
#        points = np.loadtxt(file)
#        image_points.append(vu.convert_numpy2d_to_opencv(points))
#        object_points.append(vu.convert_numpy3d_to_opencv(model))
#
#    # Generates, array of arrays
#    ids = []
#    for i in range(9):
#        ids.append(np.asarray(range(140)))
#
#    retval, camera_matrix, dist_coeffs, rvecs, tvecs = \
#        ve.mono_video_calibration_expt(ids, object_points, image_points, (1920, 1080))
#
#    print(camera_matrix)
#    print(dist_coeffs)
#    print(rvecs)
#    print(tvecs)

#def test_experimental_video_stereo_calib():
#
#   This isn't worth the bother. Takes 1.5 hours to run.
#
#    object_points, left_image_points, right_image_points, ids = \
#        load_first_stereo_data()
#
#    s_reproj, s_recon, \
#        l_c, l_d, left_rvecs, left_tvecs, \
#        r_c, r_d, right_rvecs, right_tvecs, \
#        l2r_r, l2r_t, \
#        essential, fundamental = \
#        ve.stereo_video_calibration_expt(ids,
#                                         object_points,
#                                         left_image_points,
#                                         ids,
#                                         object_points,
#                                         right_image_points,
#                                         (1920, 1080))
#    print(s_reproj)
#    print(s_recon)
#    print(l2r_r)
#    print(l2r_t)



# coding=utf-8

"""scikit-surgerycalibration sphere-fitting tests"""

import numpy
from sksurgerycalibration.algorithms import sphere_fitting

def test_fit_sphere_least_squares():
    """Test that it fits a sphere with some gaussian noise"""
    x_centre = 1.74839
    y_centre = 167.0899222
    z_centre = 200.738829

    radius = 7.543589

    #some arrays to fit data to
    coordinates = numpy.ndarray(shape=(1000, 3), dtype=float)

    #fill the arrays with points uniformly spread on
    #a sphere centred at x,y,z with radius radius
    #first seed the random generator so we get consistent test behaviour
    numpy.random.seed(seed=0)
    for i in range(1000):
        #make a random vector
        x_vector = numpy.random.uniform(-1.0, 1.0)
        y_vector = numpy.random.uniform(-1.0, 1.0)
        z_vector = numpy.random.uniform(-1.0, 1.0)

        #scale it to length radius
        length = numpy.sqrt((x_vector)**2 + (y_vector)**2 + (z_vector)**2)
        factor = radius / length

        coordinates[i, 0] = x_vector*factor + x_centre
        coordinates[i, 1] = y_vector*factor + y_centre
        coordinates[i, 2] = z_vector*factor + z_centre

    parameters = [0.0, 0.0, 0.0, 0.0]
    result = sphere_fitting.fit_sphere_least_squares(coordinates, parameters)
    numpy.testing.assert_approx_equal(result.x[0], x_centre, significant=10)
    numpy.testing.assert_approx_equal(result.x[1], y_centre, significant=10)
    numpy.testing.assert_approx_equal(result.x[2], z_centre, significant=10)
    numpy.testing.assert_approx_equal(result.x[3], radius, significant=10)



# coding=utf-8

""" Command line processing for pivot calibration. """

import argparse
from sksurgerycalibration import __version__
from sksurgerycalibration.ui.pivot_calibration_app import run_pivot_calibration


def main(args=None):
    """Entry point for pivot calibration application. """

    parser = argparse.ArgumentParser(description='pivotcalibration')

    parser.add_argument("-i", "--input",
                        required=True,
                        type=str,
                        help="A directory containing tracking matrix files.")

    parser.add_argument("-c", "--config",
                        required=False,
                        type=str,
                        help="Configuration file containing the parameters")

    version_string = __version__
    friendly_version_string = version_string if version_string else 'unknown'
    parser.add_argument(
        "-v", "--version",
        action='version',
        version='sksurgerycalibration version ' + friendly_version_string)

    args = parser.parse_args(args)

    run_pivot_calibration(args.input, args.config)
